Title:   Getting Started with Ubuntu Product Integration
Version: 20151211

Ubuntu Product Integration is a system for automated integration
testing of products built upon Ubuntu Core.

The system is designed to facilitate Product testing for both Gold
Master image releases and ongoing update testing, allowing Product
owners to control updates to devices in the field. The system will
monitor the components of your product and dispatch tests when new updates
are present, prior to those updates being made available to end users.

This tutorial will guide you step-by-step to create a snap for an application,
define a product that includes that application, create a laboratory
environment where integration tests for that product can run, and set up the
system so that it automatically executes tests in that laboratory when new
versions of the snap become available.

The document is organized as follows:

-   **[What you will need:](#what-you-will-need)** A description of everything needed to complete the tutorial.
-   **[Lab setup:](#lab-setup)** Create and configure your own laboratory.
-   **[Gold Master Candidate Testing:](#gold-master-candidate-testing)** Testing of Product images composed from the stable channel, to support releasing Gold Master Images.
-   **[Product update testing:](#product-update-testing)** Testing of software updates prior to releasing them to customers.
-   **[First-boot update testing:](#first-boot-update-testing)** First-boot update testing.

<div class="panel panel-info">
  <div class="panel-heading">
    <h3 class="panel-title">Conventions</h3>
  </div>
  <div class="panel-body">
    From time to time you will need to execute commands in a terminal. This
    will be indicated using <code>this particular font and background color</code>,
    normally on a separate line. This font will also be used to specify
    filenames or any other cases where spelling matters.
  </div>
</div>


## What you will need

In order to complete this tutorial, you will need the following resources:

-   A host computer running [Ubuntu 14.04 (64-bit)](http://www.ubuntu.com/download/desktop), with an SD card reader and at least 6 GB free drive space.
-   A [Raspberry Pi 2 Model B](https://www.raspberrypi.org/products/raspberry-pi-2-model-b/).
-   A 4GB (minimum) Micro SDHC card.
-   The host computer and Raspberry Pi connected on the same local ethernet network.
-   Access to the Internet from the host computer.
-   An optional serial USB cable, for debugging the RPI boot sequence.
-   Ubuntu One OAuth credentials. This will be covered below
    during [Lab setup](#lab-setup).

## Lab setup

In order to run tests on your product, the system needs a host computer to serve as a
bridge between your device and itself. This computer can be as simple as your
desktop, or part of a full device laboratory.

For the purposes of this tutorial, your desktop computer will serve as the
host and you will use a Raspberry Pi as the product device. The host computer
will run a small application called the Agent, which will communicate with
Canonical's servers and manage integration tests on the Raspberry Pi.

### Setting up the host computer

On your host computer, first install the latest Snappy tools:

    sudo add-apt-repository ppa:snappy-dev/tools
    sudo apt-get update
    sudo apt-get dist-upgrade
    sudo apt-get install snappy-tools

See [Snappy Getting Started Guide](https://developer.ubuntu.com/en/snappy/start/) for further details.

Also install the software dependencies of the Agent:

    sudo apt-get install python-requests python-requests-oauthlib pv

Next you will install and configure the Agent. The most recent release of the Agent can always be found at [this location](https://spi.canonical.com/assets/agent.tar.gz).

Download and install the Agent from your host computer as follows:

    wget https://spi.canonical.com/assets/agent.tar.gz
    tar -zxf agent.tar.gz
    cd agent-binaries

### Configuring the Agent

Once the Agent is installed, you will need to create its configuration.

The Agent authenticates using an Ubuntu One account . To use it, first create an
[Ubuntu One account](https://login.ubuntu.com), then instruct the Agent to put
your credentials in a configuration file which it will create with
default settings:

    ./agent.py --interactive-login config.ini

<div class="panel panel-danger">
  <div class="panel-heading">
    <h3 class="panel-title">Security warning</h3>
  </div>
  <div class="panel-body">
    The config file now has data (<code>consumer_secret</code>,
    <code>token_secret</code>, and <code>token_key</code>) that are equivalent
    to your password. You are responsible for storing this securely. The device
    on which the Agent runs should be protected from unauthorized access.
  </div>
</div>


Note: It is best practice to create a separate Ubuntu One account to use for your Agents, but we'll keep things simple for the purposes of this tutorial.

### Agent Provisioning Kit setup

When the Agent receives tests from the server, it will try to run the commands to provision, set up, and run the tests.

Together these scripts are called a "Provisioning Kit." A sample kit has been provided for this tutorial and as a starting point for your own projects.

Edit ```config.ini``` and replace the following three fields:

    provcommand = PATH_TO_AGENT/rpi2-sample-provkit/provision2
    setupcommand = PATH_TO_AGENT/rpi2-sample-provkit/setup2
    testcommand = PATH_TO_AGENT/rpi2-sample-provkit/runtest2

<div class="panel panel-warning">
  <div class="panel-heading">
    <h3 class="panel-title">Path sensitivity</h3>
  </div>
  <div class="panel-body">
     The paths in <code>config.ini</code> must be absolute. Be sure to replace
     the correct values.
  </div>
</div>

Further details of Agent configuration can be found in the [Agent Reference](agent-reference-products.html).


### Raspberry Pi Setup

Next, you will do a test provisioning to ensure your Raspberry Pi is setup properly.

Download an already prepared sample image:

    wget https://spi.canonical.com/assets/spi-test_pi2.img.xz

Verify that the MD5 hash for the downloaded file matches the value shown below:

    md5sum spi-test_pi2.img.xz
    be7cf02999e576aa5eb76e7e85713812  spi-test_pi2.img.xz

Because remotely flashing the Raspberry Pi is a feature still in active
development, you will need to manually flash an SD card with the image.

On the host machine, insert your SD card and determine which device it is
(using ```df -h``` or other tools), then copy the image to your flash device:

    xzcat spi-test_pi2.img | pv -s 4g | sudo dd of=/dev/sdX bs=32M
    sync

<div class="panel panel-danger">
  <div class="panel-heading">
    <h3 class="panel-title">Careful device selection</h3>
  </div>
  <div class="panel-body">
     Be careful select the correct output device ('of' flag). The dd command
     will offer no prompt, nor any protection from accidentally selecting your
     hard drive.
  </div>
</div>

This should take less than five minutes on most modern hardware. For additional instructions see [Getting started with a Raspberry Pi 2](https://developer.ubuntu.com/en/snappy/start/raspberry-pi-2/)

When the flashing is complete, insert the SD card in your Raspberry Pi 2, power
it on, and wait for it to boot. This typically takes less than a minute to
complete.

You can check whether it's online just by running:

    ping webdm.local

If the device is not accessible, you can troubleshoot by using a serial cable
to [obtain a console on the Raspberry Pi](http://elinux.org/RPi_Serial_Connection).
Note that the pins are the same for both versions of the Raspberry Pi.

Alternatively, it may be that the image is not properly written. If the device
fails to boot, put the SD card back in the host machine and write the image a
second time. Verify the write was succesful by confirming that the following
commands both return the same hash:

    sudo dd if=/dev/sdX | head -c `stat -c "%s" spi-test_pi2.img` | pv -s `stat -c "%s" spi-test_pi2.img` | sha1sum
    sha1sum spi-test_pi2.img

Once you are able to ping the Raspberry Pi, you can connect via ssh
(password; ubuntu):

    ssh webdm.local

And then run a snappy cli command, e.g.:

    snappy list -v


## Creating a Product

Now it is time to create your first product based on Ubuntu Core.

A special type of snap called a "Gadget Snap" (sometimes refererred to as an "OEM Snap" in the current docs) is used to define the components of a device image, i.e. the software stack that powers your device. This includes specifying the base Ubuntu OS, the specific kernel to use, pre-installed applications, configuring hardware access, etc. See [Snappy OEM Documentation](https://developer.ubuntu.com/en/snappy/guides/oem/) for more details.

Product Integration data is associated to a snap that you own, which may be either an Application or Gadget snap. 

For the purposes of this tutorial, you will use the reference ```rpi2.canonical``` gadget snap, and define your product based on an Application snap that you will own. For a real product you would create and manage your own Gadget snap as well.


### Publishing an Application Snap

We will start by having you publish a copy of our test application snap under your own MyApps account.

First, download the test snap:

    wget https://spi.canonical.com/assets/test-snap_1_all.snap

Then you need to publish the snap under your MyApps account, as follows:

- visit https://myapps.developer.ubuntu.com/dev/click-apps/upload/
- drag and drop or otherwise select the example snap for upload
- select "15.04-core" as the release
- select "Developer Tools" as the department
- select "GNU GPL v3" as the license
- click "Submit to the Ubuntu store"

If all went well, you should now see a page for the package, named like "test-snap.<username>". It will also state "Package status is Ready to Publish".

Note: when you have a separate Ubuntu One user for your Agents, this is where you would share access to the Product, by sharing this "primary snap" with the Agent user.

Now click the "Publish" button on the test-snap.<username> MyApps page. You will then see details showing that revision 1 is published in all channels.

<div class="panel panel-info">
  <div class="panel-heading">
    <h3 class="panel-title">Public Package Sharing in Development</h3>
  </div>
  <div class="panel-body">
    Note: Currently only private packages can be shared; public package
    sharing is under development. As a result, you will also need to mark
    your package as "Private" in the store.
  </div>
</div>

You should also be able to see this snap listed when you register the
agent with Canonical's servers:

    ./agent.py --register config.ini
    [agent-queue] Registered at https://spi.canonical.com/labs/agent-queue
    [agent-queue] Available products: ['test-snap.yournamespace']

If instead you see an error similar to the following, please double check your credentials and make sure you have logged into MyApps and accepted the [MyApps Terms and Conditions](https://myapps.developer.ubuntu.com):

    {
      "id": "unauthorized",
      "message": "unauthorized"
    }


### Defining the Product

So far we have published the application snap, and can build an image with it pre-installed on top of the reference Rasberry Pi Gadget snap. Now we want to codify that arrangement by defining a Product in the Ubuntu Product Integration system.

For the purposes of Product Integration testing, we are only concerned about the list of snaps that compose the product and some additional metadata decribed here:

-   ```name```: A short name to identify this Product, does not need to be unique.
-   ```primary_snap_name```: the fully qualified name of the snap
    package that defines this product and access to product test information.
-   ```release```: the Ubuntu Core release this products aims for (currently "15.04-core" or "rolling-core" as known by the Store)
-   ```snaps```: the list of snap names that that comprise the product

When you created your account in MyApps, you were required to set your "Developer Namespace". If you dont remember it, you can see it in your profile page in MyApps.
We will export an environment variable to use that namespace:

    export NAMESPACE=yournamespace

<div class="panel panel-info">
  <div class="panel-heading">
    <h3 class="panel-title">API Example Script and Authentication</h3>
  </div>
  <div class="panel-body">
    In the steps below you will be using an api_example.py script
    that is provided with the Agent distribution. Further details of
    that script and the Oauth authentication mechanism are covered in
    the <a href="api-reference-products.html">API Reference</a>.
  </div>
</div>

In the agent directory on your computer (where you already authenticated during
the [Lab setup](#lab-setup) above), issue the following HTTP request using the api_example.py helper script:

    ./api_example.py -X POST config.ini https://spi.canonical.com/products --data '{
        "name": "test-product.'$NAMESPACE'",
        "primary_snap_name": "test-snap.'$NAMESPACE'",
        "release": "15.04-core",
        "snaps": [
            "pi2.canonical",
            "webdm.canonical",
            "test-snap.'$NAMESPACE'"
        ]
    }'

    HTTP 201
    {
    "product": {
        "release": "15.04-core",
        "snaps": [
            "pi2.canonical",
            "webdm.canonical",
            "test-snap.yournamespace"
        ],
        "created_at": "2015-11-30T17:01:02.907521",
        "name": "test-product.yournamespace",
        "primary_snap_name": "test-snap.yournamespace",
        "id": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
        "active": true
        }
    }


the output will contain the ID newly created product definition. You will need to export this as an environment variable for use in later API calls as follows:

    export PRODUCT_ID=<id from output above>

<div class="panel panel-info">
  <div class="panel-heading">
    <h3 class="panel-title">A note about ubuntu-core and ubuntu-kernel snaps</h3>
  </div>
  <div class="panel-body">
    The splitting of ubuntu-core and ubuntu-kernel OS snaps is still under
  development, so no core or kernel snaps are defined as part of the product's snaps.
  </div>
</div>


## Testing

### Defining a Test

Next you will define your test scenario via a Test Specification, with the following fields:

-   ```name```: a unique reference to identify tests and their results
-   ```product_id```: the unique id for the Product to test
-   ```channels```: a list of channel pairs. 'base' is the target/live channel generally, and 'update' is where the system will look for changes to trigger update test runs (covered in a later chapter).
-   ```test_payload```: information for the test to be run; in this case we put where the tests can be downloaded, and the command to trigger their execution

Execute the following command to create this Test Specification:

    ./api_example.py -X POST config.ini https://spi.canonical.com/products/$PRODUCT_ID/tests/specs --data '{
        "name": "Tutorial Example Test Specification",
        "platform": "armhf",
        "product_id": "'$PRODUCT_ID'",
        "channel_combos": [ {"base": "stable", "update": "edge"} ],
        "test_payload": {
            "tarball_url": "https://spi.canonical.com/assets/test_tarball2.tar.gz",
            "test_cmd": "./test2"
        }
    }'

    HTTP 201
    {
        "spec": {
                "name": "Tutorial Example Test Specification",
                "id": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                "channel_combos": [
                    {
                        "base": "stable",
                        "update": "edge"
                    }
                ],
                "platform": "armhf",
                "test_payload": {
                "tarball_url": "https://spi.canonical.com/assets/test_tarball2.tar.gz",
                "test_cmd": "./test2"
            },
            "created_at": "2015-11-30T17:34:38.025749",
            "product_id": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
            "active": true
        }
    }


At this point the system is configured for your product and tests, and will supervise the Store for new versions of the snaps listed in your Product.

### First test run

This test will be what is called a Gold Master (GM) Candidate test. It is triggered by a change to any of the component snaps in the monitored ```base``` channel defined in the Test Specification (stable in this case). The server will generate a Test Opportunity that instructs the provisioning kit script to compose a new GM candidate image from the Product's snaps, using the current revisions published in the base channel, provision that image, and then run tests against it. Later chapters of the tutorial will discuss additional testing scenarios.

When we first defined our Test Specification, the system generated a
new Test Opportunity to test the initial Gold Master Candidate image,
since that combination of snaps in the ```base``` channel had not yet been tested.

Further GM tests will be automatically generated each time a component snap is updated in the ```base``` channel and are not optional.

You can now run the agent to consume and run your first test. Run the following command:

    ./agent.py -v --once config.ini

As remote flashing of the Raspberry Pi is still in active development, the
Raspberry Pi will need to be set up manually. After receiving the test to run,
the Agent will output instructions explaining how to write the composed image
to the device. Please follow those instructions and let the Agent know when
you are ready by pressing the Enter key when prompted.

After executing the test, the Agent will automatically retrieve the
results and send them back to Canonical's servers. You can query these results
in the next section, [Checking test progress and results](#checking-test-progress-and-results).


### Checking test progress and results

At any time, you can check the progress of tests generated in the
system:

    ./api_example.py config.ini \
    https://spi.canonical.com/products/$PRODUCT_ID/tests/events

This produces a large JSON document containing all the events that
relate to this product's tests, for example:

    {"event_logs": [
        {
        "updated_at": "2015-10-28T18:35:38.197000+00:00",
        "events": [
            {
            "timestamp": "2015-10-28T18:34:56.614000+00:00",
            "event_type": "QUEUED_FOR_AGENTS"
            },
            {
            "platform": "armhf",
            "timestamp": "2015-10-28T18:34:58.988000+00:00",
            "agent_group": "default",
            "agent_id": "special",
            "event_type": "PICKED_BY_AGENT"
            },

The ```event_type``` field will describe what happened, there is a timestamp, and extra information related to the event.

It will also contain all the information relevant to the test, such as:

      "queued_at": "2015-10-28T18:34:56.614000+00:00",
      "test_spec_id": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
      "spec_name": "Tutorial Example Test Specification",
      "image_unique_id": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",

As well as the test itself in the ```test_opportunity``` field.

You can also check the results for a specific image (you can get the image
id from the response of the test spec creation request):

    ./api_example.py config.ini \
    https://spi.canonical.com/products/$PRODUCT_ID/tests/events?image_unique_id=<unique_image_id>

Or you may directly poll for specific results using filters, as below:

    ./api_example.py config.ini \
    https://spi.canonical.com/products/$PRODUCT_ID/tests/results?image_unique_id=<unique_image_id>

The result query will return a document describing one or more test results.
Each will contain, among other things, the pass or failure status of the test
run, as well as all related information such as the manifest, test spec, and so
on:

    {
    "test_results": [
        {
        "created_at": "2015-10-28T18:34:56.614000+00:00",
        "result_id": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
        "test_status": "PASSED",
        "manifest": {
        ...
      "test_spec": {
        "platform": "spi-tutorial-raspi2",
        "image_name": "tutorial-raspi2-image",
        "test_payload": {
          "test_cmd": "./test",
          "tarball_url": "https://spi.canonical.com/assets/test_tarball.tar.gz"
        },
        "name": "Tutorial Example Test Specification",
        "id": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
        "created_at": "2015-10-28T16:42:42.989056",
        "active": true
      },
    ...

### Artifacts and logs

Logs containing the output and errors of the provisioning kit's scripts are always stored in ```logs/```.
Saving those logs and any artifacts the provisioning kit creates is the responsibility of your test command.

The example provisioning kit will create a tarball with the test opportunity, the test result
and all the provisioning kit logs, using this code:

    ARTIFACTS = [
        'spi_test_result.json',
        'spi_test_opportunity.json',
        'logs/provisioning.log',
        'logs/setup.log',
        'logs/testing.log'
    ]

    [...]

    artifact_dir = os.getenv('ARTIFACT_DIR', '/tmp')
    save_cmd = 'tar czvf {} {}'.format(
        os.path.join(artifact_dir, test_opp['id'] + '.tar.gz'),
        ' '.join(ARTIFACTS),
    )
    log.info('Saving artifact with %s', save_cmd)
    subprocess.check_call(save_cmd, shell=True)

Here, the tarball is stored in /tmp by default, but you can override it in the ```[environment]``` section of your ```config.ini```:

    [environment]
    ARTIFACT_DIR='/home/user/artifacts'
 
#### Generated GM Candidate images

The images that are generated by the provisioning kit may be stored for later use. For example if tests pass and you would like to release the image you will want to retrieve the binary from the Agent host.

The sample provisioning kit stores the images in a ```./run/images```
subdirector of the agent installation. You are responsible for
handling transport of the images from there to a storage system of your choice.

Note: the sample kit does not remove old images and they can quickly eat a lot of disk space.

## Product Update Testing

Now that you have defined your Product and triggered some initial
testing of Gold Master candidate images, it's time to think about the
lifecycle of your product once it is delivered to a customer.

The Ubuntu Core is designed to provide transactional updates for
both system components and applications. A major advantage of using
the Product Integration system is to enable testing of those updates
prior to releasing them into the wild.

In this chapter we will walk through setting up and performing update
testing of your application prior to releasing a new revision to end-users. We will only cover gating of updates for packages that are under your control (your application snap in this case) as the process for gating 3rd-party packages such as ubuntu-core is still under development.

To recap the state of the system from Chapter 1, you now have:

- A Product, defined with the reference RPi2 gadget snap, webdm, and
our sample application snap.
- a Test Specification, set to monitor the stable channel as "base"
and the edge channel for updates.

Since you already declared the update channel, the Product Integration system is already monitoring that channel for you. 

## Publishing a New Revision

Let's pretend you have just finished developing a new version of your application and are ready to test out the update process. 

First, download version 2 of the test snap:

    wget https://spi.canonical.com/assets/test-snap_2_all.snap

Then you need to publish this new version under your MyApps account, as follows:

- visit: https://myapps.developer.ubuntu.com/dev/click-apps/?format=snap
- click on the test-snap package name to get to the package details page
- click "Upload new version"
- drag and drop or otherwise select the example snap for upload
- under "Channels", deselect "Stable" and select "Edge"
- click "Submit to the Ubuntu store"

<div class="panel panel-info">
  <div class="panel-heading">
    <h3 class="panel-title">About Channels</h3>
  </div>
  <div class="panel-body">
  Channels are a way to compartmentalize published revisions of snap packages. Typically end-user devices will be set to use the <code>stable</code> channel and will receive only updates that are published there. By publishing first to the <code>edge</code> channel, we are able to have a new published revision that will not reach end-users but will allow us to run tests on it.  If you were to publish directly to <code>stable</code>, users would immediately start receving the update and you would not have a chance to verify that the update process works correctly.
  </div>
</div>


If all went well, you should now see both versions 1 and 2 of test-snap listed. If you don't see version 2 listed under the Edge channel, try refreshing the page. 

You now have an updated application snap published to the ```edge``` channel. 

## Running the Update Test

The Product Integration system automatically detects changes to the watched channels that are listed in the Test SPecifications. Since your TestSpec lists "edge" as the update channel, it will find the newly published version and generate a Test Opportunity simulating how an end-user update would be performed once this revision is promoted to the stable channel.

This may take a few minutes, but you can run the Agent now, without the --once flag we used earlier and it will regulary poll the system for new tests:

    ./agent.py -v config.ini

After executing the test, the Agent will automatically retrieve the
results and send them back to Canonical's servers. Test results have a type field that is one of "gold-master" or "update". As before, you can query these results using the results API endopoint, e.g.:

    ./api_example.py config.ini https://spi.canonical.com/products/$PRODUCT_ID/tests/results?image_unique_id=<unique_image_id>&type=update

You should see that the base_snaps lists the revisions that are currently present in the base channel ("stable") and update_snaps is just the new revision of test-snap that you published to the edge channel.

## Promote to Stable

Assuming the test has passed and you are ready to release this version of the application to your customers, it is now time to Promote the package via the MyApps site:

- visit: https://myapps.developer.ubuntu.com/dev/click-apps/?format=snap
- click on the test-snap package name to get to the package details page
- click on "Channels" in the left navigation
- change Stable to be revision 2
- click Update.

Note that by promoting this change to stable, the system will also generate a new Gold Master candidate test with the new set of revisions now in the stable channel.

That's it, you have now successfully gated an update until integration tests passed for the update of your application!

<div class="panel panel-info">
  <div class="panel-heading">
    <h3 class="panel-title">Gating across multiple Products</h3>
  </div>
  <div class="panel-body">
    Gating of updates is controlled at the snap package level, not on a Product basis. As a result, 
    if you have more than one Product and wish to gate the same application package independently on each Product, you will need to create two separate packages in MyApps, e.g. mysnap-product1, mysnap-product2.
  </div>
</div>


## First-boot tests

When you promote images, as said above, the system will generate new Gold Master candidate tests, creating Product Base Images that are the base for when testing Updates.

Update tests will be created always using the latest Product Base Image, but if you release images you'll need the Update tests to also be created using those released images. This is handled automatically by the system, all you need is to mark the Product Base Image as released, and the Update tests will be generated using not only the latest Product Base Image, but also all those that were ever released (until you deactivate them).

This way you can list all the Product Base Images for a given product.

    ./api_example.py -X GET config.ini https://spi.canonical.com/products/$PRODUCT_ID/baseimages

To mark any of those as released, just do:

    ./api_example.py -X PUT config.ini https://spi.canonical.com/products/$PRODUCT_ID/baseimages/<baseimage_id> --data '{
        "released": true
    }'

After this step, that Product Base Image will be used in the Update tests. The moment you decide that any Product Base Image shouldn't be used for Update tests, you can deactivate it, by doing:

    ./api_example.py -X PUT config.ini https://spi.canonical.com/products/$PRODUCT_ID/baseimages/<baseimage_id> --data '{
        "active": false
    }'
